@@ -28,6 +28,8 @@
 #include <wayfire/view.hpp>
 #include <wayfire/plugin.hpp>
 #include <wayfire/output.hpp>
+#include <wayfire/opengl.hpp>
+#include <wayfire/util/duration.hpp>
 #include <wayfire/render-manager.hpp>
 #include <wayfire/view-transform.hpp>
 #include <wayfire/per-output-plugin.hpp>
@@ -57,22 +59,41 @@ void main() {
 )";
 
 // Supplied via ipc
-//static const char *fragment_shader =
-//    R"(
-//#version 100
-//@builtin_ext@
-//@builtin@
+// static const char *fragment_shader =
+// R"(
+// #version 100
+// @builtin_ext@
+// @builtin@
 //
-//precision mediump float;
+// precision mediump float;
 //
-//varying mediump vec2 uvpos;
+// varying mediump vec2 uvpos;
 //
-//void main()
-//{
-//    vec4 c = get_pixel(uvpos);
-//    gl_FragColor = c;
-//}
-//)";
+// void main()
+// {
+// vec4 c = get_pixel(uvpos);
+// gl_FragColor = c;
+// }
+// )";
+
+static std::string pixdecor_custom_data_name = "wf-decoration-shadow-margin";
+
+class wf_shadow_margin_t : public wf::custom_data_t
+{
+  public:
+    wf::decoration_margins_t get_margins()
+    {
+        return margins;
+    }
+
+    void set_margins(wf::decoration_margins_t margins)
+    {
+        this->margins = margins;
+    }
+
+  private:
+    wf::decoration_margins_t margins = {0, 0, 0, 0};
+};
 
 namespace wf
 {
@@ -85,7 +106,9 @@ const std::string transformer_name = "filters";
 class wf_filters : public wf::scene::view_2d_transformer_t
 {
     wayfire_view view;
+    wf::output_t *output;
     OpenGL::program_t *shader;
+    std::unique_ptr<wf::animation::simple_animation_t> fade;
 
   public:
     OpenGL::program_t program;
@@ -114,8 +137,7 @@ class wf_filters : public wf::scene::view_2d_transformer_t
         }
 
         ~simple_node_render_instance_t()
-        {
-        }
+        {}
 
         void schedule_instructions(
             std::vector<render_instruction_t>& instructions,
@@ -129,12 +151,11 @@ class wf_filters : public wf::scene::view_2d_transformer_t
                         });
         }
 
-        void render(const wf::render_target_t& target,
-            const wf::region_t& region)
+        void render(const wf::scene::render_instruction_t& data)
         {
             wlr_box fb_geom =
-                target.framebuffer_box_from_geometry_box(target.geometry);
-            auto view_box = target.framebuffer_box_from_geometry_box(
+                data.target.framebuffer_box_from_geometry_box(data.target.geometry);
+            auto view_box = data.target.framebuffer_box_from_geometry_box(
                 self->get_children_bounding_box());
             view_box.x -= fb_geom.x;
             view_box.y -= fb_geom.y;
@@ -155,56 +176,121 @@ class wf_filters : public wf::scene::view_2d_transformer_t
                 0.0f, 1.0f
             };
 
-            OpenGL::render_begin(target);
-
-            /* Upload data to shader */
-            auto src_tex = wf::scene::transformer_render_instance_t<transformer_base_node_t>::get_texture(
-                1.0);
-            this->self->shader->use(src_tex.type);
-            this->self->shader->attrib_pointer("position", 2, 0, vertexData);
-            this->self->shader->attrib_pointer("texcoord", 2, 0, texCoords);
-            this->self->shader->uniformMatrix4f("mvp", target.transform);
-            this->self->shader->uniform1i("in_tex", 0);
-            GL_CALL(glActiveTexture(GL_TEXTURE0));
-            this->self->shader->set_active_texture(src_tex);
-
-            /* Render it to target */
-            target.bind();
-            GL_CALL(glViewport(x, target.viewport_height - y - h, w, h));
-
-            GL_CALL(glEnable(GL_BLEND));
-            GL_CALL(glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA));
-
-            for (const auto& box : region)
+            auto src_tex = wf::gles_texture_t{get_texture(1.0)};
+            data.pass->custom_gles_subpass(data.target,[&]
             {
-                target.logic_scissor(wlr_box_from_pixman_box(box));
-                GL_CALL(glDrawArrays(GL_TRIANGLE_FAN, 0, 4));
-            }
+                this->self->shader->use(src_tex.type);
+                this->self->shader->attrib_pointer("position", 2, 0, vertexData);
+                this->self->shader->attrib_pointer("texcoord", 2, 0, texCoords);
+                this->self->shader->uniformMatrix4f("mvp", wf::gles::output_transform(data.target));
+                this->self->shader->uniform1f("progress", *self->fade);
+                this->self->shader->uniform1i("in_tex", 0);
 
-            /* Disable stuff */
-            GL_CALL(glDisable(GL_BLEND));
-            GL_CALL(glActiveTexture(GL_TEXTURE0));
-            GL_CALL(glBindTexture(GL_TEXTURE_2D, 0));
-            GL_CALL(glBindFramebuffer(GL_FRAMEBUFFER, 0));
+                if (auto toplevel = wf::toplevel_cast(this->view))
+                {
+                    auto bg = view->get_surface_root_node()->get_bounding_box();
+                    auto vg = toplevel->get_geometry();
+                    auto margins =
+                        glm::vec4{vg.x - bg.x, vg.y - bg.y, bg.width - ((vg.x - bg.x) + vg.width),
+                        bg.height - ((vg.y - bg.y) + vg.height)};
+                    if (view->has_data(pixdecor_custom_data_name))
+                    {
+                        auto decoration_margins =
+                            view->get_data<wf_shadow_margin_t>(pixdecor_custom_data_name)->get_margins();
+                        margins.x += decoration_margins.left;
+                        margins.y += decoration_margins.bottom;
+                        margins.z += decoration_margins.right;
+                        margins.w += decoration_margins.top;
+                    }
 
-            this->self->shader->deactivate();
-            OpenGL::render_end();
-            wf::scene::damage_node(this->self, this->self->get_children_bounding_box());
+                    // XXX: Pad the margins if there are none, so that the shader renders on the surface
+                    if (bg == vg)
+                    {
+                        margins.x += 2.0;
+                        margins.y += 2.0;
+                        margins.z += 2.0;
+                        margins.w += 2.0;
+                    }
+
+                    this->self->shader->uniform4f("margins", margins);
+                }
+
+                GL_CALL(glActiveTexture(GL_TEXTURE0));
+                this->self->shader->set_active_texture(src_tex);
+
+                /* Render it to target */
+                wf::gles::bind_render_buffer(data.target);
+                GL_CALL(glViewport(x, y, w, h));
+
+                GL_CALL(glEnable(GL_BLEND));
+                GL_CALL(glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA));
+
+                for (const auto& box : data.damage)
+                {
+                    wf::gles::render_target_logic_scissor(data.target, wlr_box_from_pixman_box(box));
+                    GL_CALL(glDrawArrays(GL_TRIANGLE_FAN, 0, 4));
+                }
+
+                /* Disable stuff */
+                GL_CALL(glDisable(GL_BLEND));
+                GL_CALL(glActiveTexture(GL_TEXTURE0));
+                GL_CALL(glBindTexture(GL_TEXTURE_2D, 0));
+                GL_CALL(glBindFramebuffer(GL_FRAMEBUFFER, 0));
+
+                this->self->shader->deactivate();
+            });
         }
     };
 
     wf_filters(wayfire_view view, std::string shader_path) : wf::scene::view_2d_transformer_t(view)
     {
-        this->view    = view;
+        this->view   = view;
         this->shader = &program;
+        if (view->get_output())
+        {
+            output = view->get_output();
+            output->render->add_effect(&pre_hook, wf::OUTPUT_EFFECT_PRE);
+        }
 
         std::ifstream t(shader_path);
         std::string fragment_shader((std::istreambuf_iterator<char>(t)), std::istreambuf_iterator<char>());
-        OpenGL::render_begin();
-        program.compile(vertex_shader, fragment_shader);
-        OpenGL::render_end();
+        wf::gles::run_in_context([&]
+        {
+            program.compile(vertex_shader, fragment_shader);
+        });
+        fade = std::make_unique<wf::animation::simple_animation_t>(wf::create_option<int>(700));
+        fade->set(0.0, 0.0);
+        fade->animate(1.0);
     }
 
+    void unapply()
+    {
+        fade->animate(0.0);
+    }
+
+    void pop_transformer(wayfire_view view)
+    {
+        if (view->get_transformed_node()->get_transformer(transformer_name))
+        {
+            LOGI("Removing shader and transformer.");
+            view->get_transformed_node()->rem_transformer(transformer_name);
+        }
+    }
+
+    wf::effect_hook_t pre_hook = [=] ()
+    {
+        if (fade->running())
+        {
+            for (auto & v : wf::get_core().get_all_views())
+            {
+                v->damage();
+            }
+        } else if (fade->end == 0.0)
+        {
+            pop_transformer(view);
+        }
+    };
+
     void gen_render_instances(std::vector<render_instance_uptr>& instances,
         damage_callback push_damage, wf::output_t *shown_on) override
     {
@@ -214,14 +300,21 @@ class wf_filters : public wf::scene::view_2d_transformer_t
 
     virtual ~wf_filters()
     {
-        OpenGL::render_begin();
-        program.free_resources();
-        OpenGL::render_end();
+        wf::gles::run_in_context([&]
+        {
+            program.free_resources();
+        });
+        fade.reset();
+        if (output)
+        {
+            output->render->rem_effect(&pre_hook);
+        }
     }
 };
 
 class wayfire_per_output_filters : public wf::per_output_plugin_instance_t
 {
+    std::unique_ptr<wf::animation::simple_animation_t> fade;
     std::shared_ptr<OpenGL::program_t> program = nullptr;
     wf::post_hook_t hook;
     bool active = false;
@@ -229,20 +322,49 @@ class wayfire_per_output_filters : public wf::per_output_plugin_instance_t
   public:
     void init() override
     {
-        hook = [=] (const wf::framebuffer_t& source,
-                    const wf::framebuffer_t& destination)
+        hook = [=] (wf::auxilliary_buffer_t& aux_buf, const wf::render_buffer_t& render_buf)
         {
-            render(source, destination);
+            render(aux_buf, render_buf);
         };
+        fade = std::make_unique<wf::animation::simple_animation_t>(wf::create_option<int>(700));
+        fade->set(0.0, 0.0);
     }
 
-    nlohmann::json set_fs_shader(std::string shader)
+    wf::effect_hook_t pre_hook = [=] ()
+    {
+        if (fade->running())
+        {
+            output->render->damage_whole();
+            for (auto & v : wf::get_core().get_all_views())
+            {
+                v->damage();
+            }
+        } else if (fade->end == 0.0)
+        {
+            output->render->rem_effect(&pre_hook);
+            output->render->rem_post(&hook);
+            output->render->damage_whole();
+            if (program)
+            {
+                wf::gles::run_in_context([&]
+                {
+                    program->free_resources();
+                });
+            }
+
+            program = nullptr;
+            active  = false;
+        }
+    };
+
+    wf::json_t set_fs_shader(std::string shader)
     {
         if (program)
         {
-            OpenGL::render_begin();
-            program->free_resources();
-            OpenGL::render_end();
+            wf::gles::run_in_context([&]
+            {
+                program->free_resources();
+            });
         } else
         {
             program = std::make_shared<OpenGL::program_t>();
@@ -250,9 +372,10 @@ class wayfire_per_output_filters : public wf::per_output_plugin_instance_t
 
         std::ifstream t(shader);
         std::string fragment_shader((std::istreambuf_iterator<char>(t)), std::istreambuf_iterator<char>());
-        OpenGL::render_begin();
-        program->compile(vertex_shader, fragment_shader);
-        OpenGL::render_end();
+        wf::gles::run_in_context([&]
+        {
+            program->compile(vertex_shader, fragment_shader);
+        });
         if (program->get_program_id(wf::TEXTURE_TYPE_RGBA) == 0)
         {
             LOGE("Failed to compile fullscreen shader.");
@@ -260,6 +383,7 @@ class wayfire_per_output_filters : public wf::per_output_plugin_instance_t
             program = nullptr;
             return wf::ipc::json_error("Failed to compile fullscreen shader.");
         }
+
         output->render->damage_whole();
 
         if (active)
@@ -269,36 +393,28 @@ class wayfire_per_output_filters : public wf::per_output_plugin_instance_t
         }
 
         output->render->add_post(&hook);
+        output->render->add_effect(&pre_hook, wf::OUTPUT_EFFECT_PRE);
+        fade->animate(1.0);
         active = true;
 
         LOGI("Successfully compiled and applied fullscreen shader to output: ", output->to_string());
         return wf::ipc::json_ok();
-    };
+    }
 
-    nlohmann::json unset_fs_shader()
+    wf::json_t unset_fs_shader()
     {
-        output->render->rem_post(&hook);
-        output->render->damage_whole();
-        if (program)
-        {
-            OpenGL::render_begin();
-            program->free_resources();
-            OpenGL::render_end();
-        }
-        program = nullptr;
-        active = false;
+        fade->animate(0.0);
         return wf::ipc::json_ok();
-    };
+    }
 
-    nlohmann::json fs_has_shader()
+    wf::json_t fs_has_shader()
     {
         auto response = wf::ipc::json_ok();
         response["has-shader"] = active;
         return response;
-    };
+    }
 
-    void render(const wf::framebuffer_t& source,
-        const wf::framebuffer_t& target)
+    void render(wf::auxilliary_buffer_t& aux_buf, const wf::render_buffer_t& render_buf)
     {
         static const float vertexData[] = {
             -1.0f, -1.0f,
@@ -307,59 +423,63 @@ class wayfire_per_output_filters : public wf::per_output_plugin_instance_t
             -1.0f, 1.0f
         };
         static const float texCoords[] = {
-            0.0f, 0.0f,
-            1.0f, 0.0f,
+            0.0f, 1.0f,
             1.0f, 1.0f,
-            0.0f, 1.0f
+            1.0f, 0.0f,
+            0.0f, 0.0f
         };
 
-        OpenGL::render_begin(target);
+        wf::gles::run_in_context([&]
+        {
+            /* Upload data to shader */
+            program->use(wf::TEXTURE_TYPE_RGBA);
+            program->attrib_pointer("position", 2, 0, vertexData);
+            program->attrib_pointer("texcoord", 2, 0, texCoords);
+            program->uniformMatrix4f("mvp", glm::mat4(1.0));
+            program->uniform1f("progress", *fade);
+            program->uniform1i("in_tex", 0);
+            GL_CALL(glActiveTexture(GL_TEXTURE0));
+            program->set_active_texture(wf::gles_texture_t::from_aux(aux_buf));
 
-        /* Upload data to shader */
-        program->use(wf::TEXTURE_TYPE_RGBA);
-        program->attrib_pointer("position", 2, 0, vertexData);
-        program->attrib_pointer("texcoord", 2, 0, texCoords);
-        program->uniformMatrix4f("mvp", glm::mat4(1.0));
-        program->uniform1i("in_tex", 0);
-        GL_CALL(glActiveTexture(GL_TEXTURE0));
-        program->set_active_texture(source.tex);
+            /* Render it to aux_buf */
+            wf::gles::bind_render_buffer(render_buf);
+            GL_CALL(glViewport(0, 0, aux_buf.get_size().width, aux_buf.get_size().height));
 
-        /* Render it to target */
-        target.bind();
-        GL_CALL(glViewport(0, 0, target.viewport_width, target.viewport_height));
+            GL_CALL(glEnable(GL_BLEND));
+            GL_CALL(glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA));
 
-        GL_CALL(glEnable(GL_BLEND));
-        GL_CALL(glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA));
+            GL_CALL(glDrawArrays(GL_TRIANGLE_FAN, 0, 4));
 
-        GL_CALL(glDrawArrays(GL_TRIANGLE_FAN, 0, 4));
+            /* Disable stuff */
+            GL_CALL(glDisable(GL_BLEND));
+            GL_CALL(glActiveTexture(GL_TEXTURE0));
+            GL_CALL(glBindTexture(GL_TEXTURE_2D, 0));
+            GL_CALL(glBindFramebuffer(GL_FRAMEBUFFER, 0));
 
-        /* Disable stuff */
-        GL_CALL(glDisable(GL_BLEND));
-        GL_CALL(glActiveTexture(GL_TEXTURE0));
-        GL_CALL(glBindTexture(GL_TEXTURE_2D, 0));
-        GL_CALL(glBindFramebuffer(GL_FRAMEBUFFER, 0));
-
-        program->deactivate();
-        OpenGL::render_end();
+            program->deactivate();
+        });
     }
 
     void fini() override
     {
+        output->render->rem_effect(&pre_hook);
         output->render->rem_post(&hook);
         output->render->damage_whole();
         if (program)
         {
-            OpenGL::render_begin();
-            program->free_resources();
-            OpenGL::render_end();
+            wf::gles::run_in_context([&]
+            {
+                program->free_resources();
+            });
         }
+
+        fade.reset();
     }
 };
 
 class wayfire_filters : public wf::plugin_interface_t,
     public wf::per_output_tracker_mixin_t<wayfire_per_output_filters>
 {
-    std::map<wayfire_view, std::shared_ptr<wf_filters>> transformers;
     wf::shared_data::ref_ptr_t<wf::ipc::method_repository_t> ipc_repo;
 
     void pop_transformer(wayfire_view view)
@@ -405,9 +525,9 @@ class wayfire_filters : public wf::plugin_interface_t,
     std::shared_ptr<wf_filters> ensure_transformer(wayfire_view view, std::string shader_path)
     {
         auto tmgr = view->get_transformed_node();
-        if (tmgr->get_transformer<wf::scene::node_t>(transformer_name))
+        if (tmgr->get_transformer<wf_filters>(transformer_name))
         {
-            pop_transformer(view);
+            view->get_transformed_node()->rem_transformer(transformer_name);
         }
 
         auto node = std::make_shared<wf_filters>(view, shader_path);
@@ -416,16 +536,16 @@ class wayfire_filters : public wf::plugin_interface_t,
         return tmgr->get_transformer<wf_filters>(transformer_name);
     }
 
-    wf::ipc::method_callback ipc_set_view_shader = [=] (nlohmann::json data) -> nlohmann::json
+    wf::ipc::method_callback ipc_set_view_shader = [=] (wf::json_t data) -> wf::json_t
     {
-        WFJSON_EXPECT_FIELD(data, "view-id", number_unsigned);
-        WFJSON_EXPECT_FIELD(data, "shader-path", string);
+        auto view_id     = wf::ipc::json_get_uint64(data, "view-id");
+        auto shader_path = wf::ipc::json_get_string(data, "shader-path");
 
-        auto view = wf::ipc::find_view_by_id(data["view-id"]);
-        if (view && view->is_mapped())
+        auto view = wf::ipc::find_view_by_id(view_id);
+        if (view)
         {
-            transformers[view] = ensure_transformer(view, data["shader-path"]);
-            if (transformers[view]->program.get_program_id(wf::TEXTURE_TYPE_RGBA) == 0)
+            auto tr = ensure_transformer(view, shader_path);
+            if (tr->program.get_program_id(wf::TEXTURE_TYPE_RGBA) == 0)
             {
                 pop_transformer(view);
                 LOGE("Failed to compile shader.");
@@ -442,30 +562,35 @@ class wayfire_filters : public wf::plugin_interface_t,
         return wf::ipc::json_ok();
     };
 
-    wf::ipc::method_callback ipc_unset_view_shader = [=] (nlohmann::json data) -> nlohmann::json
+    wf::ipc::method_callback ipc_unset_view_shader = [=] (wf::json_t data) -> wf::json_t
     {
-        WFJSON_EXPECT_FIELD(data, "view-id", number_unsigned);
+        auto view_id = wf::ipc::json_get_uint64(data, "view-id");
 
-        auto view = wf::ipc::find_view_by_id(data["view-id"]);
+        auto view = wf::ipc::find_view_by_id(view_id);
         if (view)
         {
-            pop_transformer(view);
-            view->damage();
+            auto tmgr = view->get_transformed_node();
+            if (auto tr = tmgr->get_transformer<wf_filters>(transformer_name))
+            {
+                tr->unapply();
+                view->damage();
+            }
         }
 
         return wf::ipc::json_ok();
     };
 
-    wf::ipc::method_callback ipc_view_has_shader = [=] (nlohmann::json data) -> nlohmann::json
+    wf::ipc::method_callback ipc_view_has_shader = [=] (wf::json_t data) -> wf::json_t
     {
-        WFJSON_EXPECT_FIELD(data, "view-id", number_unsigned);
+        auto view_id = wf::ipc::json_get_uint64(data, "view-id");
 
-        auto view = wf::ipc::find_view_by_id(data["view-id"]);
+        auto view = wf::ipc::find_view_by_id(view_id);
         if (!view)
         {
             return wf::ipc::json_error("Failed to find view with given id.");
         }
-        auto tmgr = view->get_transformed_node();
+
+        auto tmgr     = view->get_transformed_node();
         auto response = wf::ipc::json_ok();
         response["has-shader"] = tmgr->get_transformer<wf::scene::node_t>(transformer_name) ? true : false;
         return response;
@@ -473,35 +598,36 @@ class wayfire_filters : public wf::plugin_interface_t,
 
     wf::output_t *find_output_by_name(std::string name)
     {
-        for (auto &output : wf::get_core().output_layout->get_outputs())
+        for (auto & output : wf::get_core().output_layout->get_outputs())
         {
             if (output->to_string() == name)
             {
                 return output;
             }
         }
+
         return nullptr;
     }
 
-    wf::ipc::method_callback ipc_set_fs_shader = [=] (nlohmann::json data) -> nlohmann::json
+    wf::ipc::method_callback ipc_set_fs_shader = [=] (wf::json_t data) -> wf::json_t
     {
-        WFJSON_EXPECT_FIELD(data, "output-name", string);
-        WFJSON_EXPECT_FIELD(data, "shader-path", string);
+        auto output_name = wf::ipc::json_get_string(data, "output-name");
+        auto shader_path = wf::ipc::json_get_string(data, "shader-path");
 
-        auto output = find_output_by_name(data["output-name"]);
+        auto output = find_output_by_name(output_name);
         if (!output)
         {
             return wf::ipc::json_error("No such output");
         }
 
-        return this->output_instance[output]->set_fs_shader(data["shader-path"]);
+        return this->output_instance[output]->set_fs_shader(shader_path);
     };
 
-    wf::ipc::method_callback ipc_unset_fs_shader = [=] (nlohmann::json data) -> nlohmann::json
+    wf::ipc::method_callback ipc_unset_fs_shader = [=] (wf::json_t data) -> wf::json_t
     {
-        WFJSON_EXPECT_FIELD(data, "output-name", string);
+        auto output_name = wf::ipc::json_get_string(data, "output-name");
 
-        auto output = find_output_by_name(data["output-name"]);
+        auto output = find_output_by_name(output_name);
         if (!output)
         {
             return wf::ipc::json_error("No such output");
@@ -510,11 +636,11 @@ class wayfire_filters : public wf::plugin_interface_t,
         return this->output_instance[output]->unset_fs_shader();
     };
 
-    wf::ipc::method_callback ipc_fs_has_shader = [=] (nlohmann::json data) -> nlohmann::json
+    wf::ipc::method_callback ipc_fs_has_shader = [=] (wf::json_t data) -> wf::json_t
     {
-        WFJSON_EXPECT_FIELD(data, "output-name", string);
+        auto output_name = wf::ipc::json_get_string(data, "output-name");
 
-        auto output = find_output_by_name(data["output-name"]);
+        auto output = find_output_by_name(output_name);
         if (!output)
         {
             return wf::ipc::json_error("No such output");
