@@ -26,10 +26,11 @@ decoration_area_t::decoration_area_t(decoration_area_type_t type, wf::geometry_t
  */
 decoration_area_t::decoration_area_t(wf::geometry_t g,
     std::function<void(wlr_box)> damage_callback,
-    const decoration_theme_t& theme)
+    pixdecor_theme_t& theme)
 {
     this->type     = DECORATION_AREA_BUTTON;
     this->geometry = g;
+    this->damage_callback = damage_callback;
 
     this->button = std::make_unique<button_t>(theme,
         std::bind(damage_callback, g));
@@ -40,6 +41,17 @@ wf::geometry_t decoration_area_t::get_geometry() const
     return geometry;
 }
 
+void decoration_area_t::set_geometry(wf::geometry_t g)
+{
+    geometry = g;
+
+    if (this->type == DECORATION_AREA_BUTTON)
+    {
+        this->button = std::make_unique<button_t>(this->button->theme,
+            std::bind(this->damage_callback, g));
+    }
+}
+
 button_t& decoration_area_t::as_button()
 {
     assert(button);
@@ -52,26 +64,26 @@ decoration_area_type_t decoration_area_t::get_type() const
     return type;
 }
 
-decoration_layout_t::decoration_layout_t(const decoration_theme_t& th,
+pixdecor_layout_t::pixdecor_layout_t(pixdecor_theme_t& th,
     std::function<void(wlr_box)> callback) :
 
     titlebar_size(th.get_title_height()),
     border_size(th.get_input_size()),
-    button_width(th.get_font_height_px() >= LARGE_ICON_THRESHOLD ? 26 : 18),
-    button_height(th.get_font_height_px() >= LARGE_ICON_THRESHOLD ? 26 : 18),
-    button_padding((titlebar_size - button_height) / 2),
     theme(th),
     damage_callback(callback)
 {}
 
-decoration_layout_t::~decoration_layout_t()
+pixdecor_layout_t::~pixdecor_layout_t()
 {
     this->layout_areas.clear();
 }
 
-wf::geometry_t decoration_layout_t::create_buttons(int width, int radius)
+wf::geometry_t pixdecor_layout_t::create_buttons(int width, int radius)
 {
     // read the string from settings; start at the colon and replace commas with spaces
+    wf::option_wrapper_t<int> button_spacing{"pixdecor/button_spacing"};
+    wf::option_wrapper_t<int> button_x_offset{"pixdecor/button_x_offset"};
+    wf::option_wrapper_t<int> button_y_offset{"pixdecor/button_y_offset"};
     GSettings *settings = g_settings_new("org.gnome.desktop.wm.preferences");
     gchar *b_layout     = g_settings_get_string(settings, "button-layout");
     gchar *ptr = b_layout + strlen(b_layout) - 1;
@@ -115,33 +127,38 @@ wf::geometry_t decoration_layout_t::create_buttons(int width, int radius)
         }
     }
 
-    int per_button = 2 * BUTTON_W_PAD + button_width;
-    int border     = theme.get_border_size();
-    wf::geometry_t button_geometry = {
-        width - (maximized ? 0 : border),
-        button_padding + border / 2 + (radius * 2),
-        button_width,
-        button_height,
-    };
+    int total_width = 0;
+    int per_button  = 0;
+    int border = theme.get_border_size();
+    wf::geometry_t button_geometry;
+    button_geometry.x = (width - (maximized ? 4 : border)) + button_x_offset;
 
     for (auto type : wf::reverse(buttons))
     {
+        auto button_area  = std::make_unique<decoration_area_t>(button_geometry, damage_callback, theme);
+        auto surface_size = button_area->as_button().set_button_type(type);
+        button_geometry.width  = surface_size.width;
+        button_geometry.height = surface_size.height;
+        int button_padding = (theme.get_title_height() - button_geometry.height) / 2 + button_y_offset;
+        button_geometry.y = button_padding + border / 2 + (radius * 2);
+        per_button = button_geometry.width + (buttons.back() == type ? 0 : button_spacing);
         button_geometry.x -= per_button;
-        this->layout_areas.push_back(std::make_unique<decoration_area_t>(
-            button_geometry, damage_callback, theme));
-        this->layout_areas.back()->as_button().set_button_type(type);
+        button_area->set_geometry(button_geometry);
+        button_area->as_button().set_button_type(type);
+        this->layout_areas.push_back(std::move(button_area));
+        total_width += per_button;
     }
 
-    int total_width = buttons.size() * per_button;
+    total_width -= button_x_offset;
 
     return {
-        button_geometry.x, maximized ? 0 : border + (radius * 2),
-        total_width, titlebar_size
+        button_geometry.x, maximized ? 4 : border + (radius * 2),
+        total_width, theme.get_title_height()
     };
 }
 
 /** Regenerate layout using the new size */
-void decoration_layout_t::resize(int width, int height)
+void pixdecor_layout_t::resize(int width, int height)
 {
     wf::option_wrapper_t<int> shadow_radius{"pixdecor/shadow_radius"};
     wf::option_wrapper_t<std::string> overlay_engine{"pixdecor/overlay_engine"};
@@ -153,7 +170,7 @@ void decoration_layout_t::resize(int width, int height)
 
     this->layout_areas.clear();
 
-    if (this->titlebar_size > 0)
+    if (this->theme.get_title_height() > 0)
     {
         auto button_geometry_expanded = create_buttons(width - (radius * 2), radius);
 
@@ -167,8 +184,8 @@ void decoration_layout_t::resize(int width, int height)
             maximized ? 0 : border / 2 + (radius * 2),
             /* Up to the button, but subtract the padding to the left of the
              * title and the padding between title and button */
-            button_geometry_expanded.x - border,
-            titlebar_size + (maximized ? 0 : border / 2 + 1),
+            std::max(1, button_geometry_expanded.x - border),
+            theme.get_title_height() + (maximized ? 0 : border / 2 + 1),
         };
         this->layout_areas.push_back(std::make_unique<decoration_area_t>(
             DECORATION_AREA_TITLE, title_geometry));
@@ -241,7 +258,7 @@ void decoration_layout_t::resize(int width, int height)
  * @return The decoration areas which need to be rendered, in top to bottom
  *  order.
  */
-std::vector<nonstd::observer_ptr<decoration_area_t>> decoration_layout_t::get_renderable_areas()
+std::vector<nonstd::observer_ptr<decoration_area_t>> pixdecor_layout_t::get_renderable_areas()
 {
     std::vector<nonstd::observer_ptr<decoration_area_t>> renderable;
     for (auto& area : layout_areas)
@@ -255,7 +272,7 @@ std::vector<nonstd::observer_ptr<decoration_area_t>> decoration_layout_t::get_re
     return renderable;
 }
 
-wf::region_t decoration_layout_t::calculate_region() const
+wf::region_t pixdecor_layout_t::calculate_region() const
 {
     wf::region_t r{};
     for (auto& area : layout_areas)
@@ -292,13 +309,13 @@ wf::region_t decoration_layout_t::calculate_region() const
     return r;
 }
 
-wf::region_t decoration_layout_t::limit_region(wf::region_t & region) const
+wf::region_t pixdecor_layout_t::limit_region(wf::region_t & region) const
 {
     wf::region_t out = region & this->cached_titlebar;
     return out;
 }
 
-void decoration_layout_t::unset_hover(wf::point_t position)
+void pixdecor_layout_t::unset_hover(wf::point_t position)
 {
     auto area = find_area_at(position);
     if (area && (area->get_type() == DECORATION_AREA_BUTTON))
@@ -308,7 +325,7 @@ void decoration_layout_t::unset_hover(wf::point_t position)
 }
 
 /** Handle motion event to (x, y) relative to the decoration */
-decoration_layout_t::action_response_t decoration_layout_t::handle_motion(
+pixdecor_layout_t::action_response_t pixdecor_layout_t::handle_motion(
     int x, int y)
 {
     auto previous_area = find_area_at(current_input);
@@ -344,7 +361,7 @@ decoration_layout_t::action_response_t decoration_layout_t::handle_motion(
  * @return The action which needs to be carried out in response to this
  *  event.
  * */
-decoration_layout_t::action_response_t decoration_layout_t::handle_press_event(
+pixdecor_layout_t::action_response_t pixdecor_layout_t::handle_press_event(
     bool pressed)
 {
     if (pressed)
@@ -411,11 +428,23 @@ decoration_layout_t::action_response_t decoration_layout_t::handle_press_event(
     return {DECORATION_ACTION_NONE, 0};
 }
 
+pixdecor_layout_t::action_response_t pixdecor_layout_t::handle_axis_event(
+    int delta)
+{
+    if (delta < 0)
+    {
+        return {DECORATION_ACTION_SHADE, 0};
+    } else
+    {
+        return {DECORATION_ACTION_UNSHADE, 0};
+    }
+}
+
 /**
  * Find the layout area at the given coordinates, if any
  * @return The layout area or null on failure
  */
-nonstd::observer_ptr<decoration_area_t> decoration_layout_t::find_area_at(
+nonstd::observer_ptr<decoration_area_t> pixdecor_layout_t::find_area_at(
     wf::point_t point)
 {
     for (auto& area : this->layout_areas)
@@ -478,7 +507,7 @@ nonstd::observer_ptr<decoration_area_t> decoration_layout_t::find_area_at(
 }
 
 /** Calculate resize edges based on @current_input */
-uint32_t decoration_layout_t::calculate_resize_edges() const
+uint32_t pixdecor_layout_t::calculate_resize_edges() const
 {
     wf::option_wrapper_t<int> shadow_radius{"pixdecor/shadow_radius"};
     wf::option_wrapper_t<std::string> overlay_engine{"pixdecor/overlay_engine"};
@@ -524,7 +553,7 @@ uint32_t decoration_layout_t::calculate_resize_edges() const
 }
 
 /** Update the cursor based on @current_input */
-void decoration_layout_t::update_cursor()
+void pixdecor_layout_t::update_cursor()
 {
     uint32_t edges = calculate_resize_edges();
     auto area = find_area_at(this->current_input);
@@ -539,12 +568,12 @@ void decoration_layout_t::update_cursor()
     wf::get_core().set_cursor(cursor_name);
 }
 
-void decoration_layout_t::set_maximize(bool state)
+void pixdecor_layout_t::set_maximize(bool state)
 {
     maximized = state;
 }
 
-void decoration_layout_t::handle_focus_lost()
+void pixdecor_layout_t::handle_focus_lost()
 {
     if (is_grabbed)
     {
