@@ -15,7 +15,6 @@
 #include <wayfire/output.hpp>
 #include <wayfire/opengl.hpp>
 #include <wayfire/core.hpp>
-#include <wayfire/view-transform.hpp>
 #include <wayfire/signal-definitions.hpp>
 #include <wayfire/toplevel-view.hpp>
 #include "deco-subsurface.hpp"
@@ -24,18 +23,29 @@
 #include <wayfire/window-manager.hpp>
 #include <wayfire/view-transform.hpp>
 #include <wayfire/txn/transaction-manager.hpp>
+#include <wayfire/scene-render.hpp>
 
 #include <wayfire/plugins/common/cairo-util.hpp>
 
 #include <cairo.h>
+#include "shade.hpp"
 
-wf::option_wrapper_t<std::string> overlay_engine{"pixdecor/overlay_engine"};
-wf::option_wrapper_t<std::string> effect_type{"pixdecor/effect_type"};
+
+namespace wf
+{
+namespace pixdecor
+{
 wf::option_wrapper_t<wf::color_t> effect_color{"pixdecor/effect_color"};
 wf::option_wrapper_t<int> shadow_radius{"pixdecor/shadow_radius"};
 wf::option_wrapper_t<bool> titlebar_opt{"pixdecor/titlebar"};
+wf::option_wrapper_t<int> csd_titlebar_height{"pixdecor/csd_titlebar_height"};
+wf::option_wrapper_t<bool> enable_shade{"pixdecor/enable_shade"};
+wf::option_wrapper_t<std::string> title_font{"pixdecor/title_font"};
+wf::option_wrapper_t<std::string> overlay_engine{"pixdecor/overlay_engine"};
+wf::option_wrapper_t<std::string> effect_type{"pixdecor/effect_type"};
 wf::option_wrapper_t<bool> maximized_borders{"pixdecor/maximized_borders"};
 wf::option_wrapper_t<bool> maximized_shadows{"pixdecor/maximized_shadows"};
+wf::option_wrapper_t<int> title_text_align{"pixdecor/title_text_align"};
 
 class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_interaction_t,
     public wf::touch_interaction_t
@@ -50,23 +60,27 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
         }
     };
 
-    void update_title(int width, int height, int t_width, double scale)
+    void update_title(int width, int height, int t_width, int border, int buttons_width, double scale)
     {
         if (auto view = _view.lock())
         {
             int target_width  = width * scale;
             int target_height = height * scale;
 
-            if ((view->get_title() != title_texture.current_text) ||
-                (target_width != title_texture.tex.width) ||
-                (target_height != title_texture.tex.height) ||
+            if ((int(title_text_align) != title_texture.title_text_align) ||
+                (view->get_title() != title_texture.current_text) ||
+                (target_width != title_texture.tex.get_size().width) ||
+                (std::string(title_font) != title_texture.title_font_string) ||
+                (target_height != title_texture.tex.get_size().height) ||
                 (view->activated != title_texture.rendered_for_activated_state))
             {
                 auto surface = theme.render_text(view->get_title(),
-                    target_width, target_height, t_width, view->activated);
-                cairo_surface_upload_to_texture(surface, title_texture.tex);
+                    target_width, target_height, t_width, border, buttons_width, view->activated);
+                title_texture.tex = owned_texture_t{surface};
                 cairo_surface_destroy(surface);
-                title_texture.current_text = view->get_title();
+                title_texture.title_font_string = title_font;
+                title_texture.current_text     = view->get_title();
+                title_texture.title_text_align = int(title_text_align);
                 title_texture.rendered_for_activated_state = view->activated;
             }
         }
@@ -74,14 +88,16 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
 
     struct
     {
-        wf::simple_texture_t tex;
+        wf::owned_texture_t tex;
         std::string current_text = "";
         bool rendered_for_activated_state = false;
+        int title_text_align = int(title_text_align);
+        std::string title_font_string = title_font;
     } title_texture;
 
   public:
-    wf::pixdecor::decoration_theme_t theme;
-    wf::pixdecor::decoration_layout_t layout;
+    pixdecor_theme_t theme;
+    pixdecor_layout_t layout;
     wf::region_t cached_region;
 
     wf::dimensions_t size;
@@ -105,7 +121,9 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
     }
 
     ~simple_decoration_node_t()
-    {}
+    {
+        remove_shade_transformers();
+    }
 
     wf::point_t get_offset()
     {
@@ -118,24 +136,26 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
         return {-current_thickness, -current_titlebar};
     }
 
-    void render_title(const wf::render_target_t& fb, const wf::region_t& scissor,
-        const wf::geometry_t& geometry, int t_width)
+    void render_title(const wf::scene::render_instruction_t& data,
+        const wf::geometry_t& geometry, int t_width, int border, int buttons_width)
     {
-        update_title(geometry.width, geometry.height, t_width, fb.scale);
-        OpenGL::render_texture(title_texture.tex.tex, fb, geometry,
-            glm::vec4(1.0f), OpenGL::TEXTURE_TRANSFORM_INVERT_Y | OpenGL::RENDER_FLAG_CACHED);
+        update_title(geometry.width, geometry.height, t_width, border, buttons_width, data.target.scale);
+        OpenGL::render_texture(wf::gles_texture_t{title_texture.tex.get_texture()}, data.target, geometry,
+            glm::vec4(1.0f), OpenGL::RENDER_FLAG_CACHED);
 
-        for (auto& box : scissor)
+        data.pass->custom_gles_subpass(data.target,[&]
         {
-            fb.logic_scissor(wlr_box_from_pixman_box(box));
-            OpenGL::draw_cached();
-        }
+            for (auto& box : data.damage)
+            {
+                wf::gles::render_target_logic_scissor(data.target, wlr_box_from_pixman_box(box));
+                OpenGL::draw_cached();
+            }
+        });
 
         OpenGL::clear_cached();
-        OpenGL::render_end();
     }
 
-    void render_region(const wf::render_target_t& fb, wf::point_t origin, const wf::region_t& region)
+    void render_region(const wf::scene::render_instruction_t& data, wf::point_t origin)
     {
         int border = theme.get_border_size();
         wlr_box geometry{origin.x, origin.y, size.width, size.height};
@@ -149,32 +169,47 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
         }
 
         auto renderables = layout.get_renderable_areas();
-        auto offset = wf::point_t{origin.x, origin.y - (maximized ? -border / 2 : border / 4)};
+        auto offset =
+            wf::point_t{origin.x,
+            origin.y -
+            ((maximized && (!maximized_shadows || !maximized_borders)) ? -border / 2 : border / 4)};
 
-        OpenGL::render_begin(fb);
-
-        theme.render_background(fb, geometry, region, activated, current_cursor_position);
-
-        if (!titlebar_opt)
+        wf::gles::run_in_context([&]
         {
-            OpenGL::render_end();
-            return;
-        }
+            wf::gles::bind_render_buffer(data.target);
 
-        /* Draw title & buttons */
-        for (auto item : renderables)
-        {
-            if (item->get_type() == wf::pixdecor::DECORATION_AREA_TITLE)
+            theme.render_background(data, geometry, activated, current_cursor_position);
+
+            if (!titlebar_opt)
             {
-                render_title(fb, region, item->get_geometry() + offset, size.width - border * 2);
-            } else // button
-            {
-                item->as_button().render(fb,
-                    item->get_geometry() + origin, region);
+                return;
             }
-        }
 
-        OpenGL::render_end();
+            int buttons_width = 0;
+            for (auto item : renderables)
+            {
+                if (item->get_type() != DECORATION_AREA_TITLE)
+                {
+                    buttons_width += item->get_geometry().width;
+                }
+            }
+
+            /* Draw title & buttons */
+            auto title_border = border + ((std::string(overlay_engine) == "rounded_corners" &&
+                (!maximized || maximized_shadows)) ? int(shadow_radius) * 2 : 0);
+            for (auto item : renderables)
+            {
+                if (item->get_type() == DECORATION_AREA_TITLE)
+                {
+                    render_title(data,
+                        item->get_geometry() + offset, size.width - border * 2, title_border, buttons_width);
+                } else // button
+                {
+                    item->as_button().render(data,
+                        item->get_geometry() + origin);
+                }
+            }
+        });
     }
 
     std::optional<wf::scene::input_node_t> find_node_at(const wf::pointf_t& at) override
@@ -186,7 +221,9 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
         }
 
         int border = theme.get_border_size();
-        int r = (std::string(overlay_engine) == "rounded_corners" && !maximized) ? int(shadow_radius) * 2 : 0;
+        int r =
+            (std::string(overlay_engine) == "rounded_corners" &&
+                (!maximized || (maximized && maximized_shadows))) ? int(shadow_radius) * 2 : 0;
         r -= MIN_RESIZE_HANDLE_SIZE - std::min(border, MIN_RESIZE_HANDLE_SIZE);
         wf::pointf_t local = at - wf::pointf_t{get_offset()};
         if (auto view = _view.lock())
@@ -194,7 +231,12 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
             wf::geometry_t g = view->get_geometry();
             g.x = g.y = 0;
             g   = wf::expand_geometry_by_margins(g, wf::decoration_margins_t{-r, -r, -r, -r});
-            if (wf::region_t{g}.contains_pointf(local))
+            wf::region_t deco_region{g};
+            g = wf::expand_geometry_by_margins(g, wf::decoration_margins_t{-border, -border, -border,
+                -theme.get_title_height() - border});
+            wf::region_t view_region{g};
+            deco_region ^= view_region;
+            if (deco_region.contains_pointf(local))
             {
                 return wf::scene::input_node_t{
                     .node = this,
@@ -243,15 +285,14 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
             if (!our_damage.empty())
             {
                 instructions.push_back(wf::scene::render_instruction_t{
-                    .instance = this,
-                    .target   = target,
-                    .damage   = std::move(our_damage),
-                });
+                            .instance = this,
+                            .target   = target,
+                            .damage   = std::move(our_damage),
+                        });
             }
         }
 
-        void render(const wf::render_target_t& target,
-            const wf::region_t& region) override
+        void render(const wf::scene::render_instruction_t& data) override
         {
             auto offset = self->get_offset();
             wlr_box rectangle{offset.x, offset.y, self->size.width, self->size.height};
@@ -265,13 +306,13 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
 
             if ((std::string(effect_type) != "none") || (std::string(overlay_engine) != "none"))
             {
-                self->theme.smoke.step_effect(target, rectangle, std::string(effect_type) == "ink",
+                self->theme.smoke.step_effect(data, rectangle, std::string(effect_type) == "ink",
                     self->current_cursor_position, self->theme.get_decor_color(activated), effect_color,
                     self->theme.get_title_height(), self->theme.get_border_size(),
                     (std::string(overlay_engine) == "rounded_corners" && !maximized) ? shadow_radius : 0);
             }
 
-            self->render_region(target, offset, region);
+            self->render_region(data, offset);
         }
     };
 
@@ -313,25 +354,91 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
             return;
         }
 
-        handle_action(layout.handle_press_event(ev.state == WLR_BUTTON_PRESSED));
+        handle_action(layout.handle_press_event(ev.state == WL_POINTER_BUTTON_STATE_PRESSED));
     }
 
-    void handle_action(wf::pixdecor::decoration_layout_t::action_response_t action)
+    void handle_pointer_axis(const wlr_pointer_axis_event& ev) override
+    {
+        if (ev.orientation == WL_POINTER_AXIS_VERTICAL_SCROLL)
+        {
+            handle_action(layout.handle_axis_event(ev.delta));
+        }
+    }
+
+    void pop_transformer(wayfire_view view)
+    {
+        if (view->get_transformed_node()->get_transformer(shade_transformer_name))
+        {
+            view->get_transformed_node()->rem_transformer(shade_transformer_name);
+        }
+    }
+
+    void remove_shade_transformers()
+    {
+        for (auto& view : wf::get_core().get_all_views())
+        {
+            pop_transformer(view);
+        }
+    }
+
+    std::shared_ptr<pixdecor_shade> ensure_transformer(wayfire_view view, int titlebar_height)
+    {
+        auto tmgr = view->get_transformed_node();
+        if (auto tr = tmgr->get_transformer<pixdecor_shade>(shade_transformer_name))
+        {
+            return tr;
+        }
+
+        auto node = std::make_shared<pixdecor_shade>(view, titlebar_height);
+        tmgr->add_transformer(node, wf::TRANSFORMER_2D, shade_transformer_name);
+        auto tr = tmgr->get_transformer<pixdecor_shade>(shade_transformer_name);
+
+        return tr;
+    }
+
+    void init_shade(wayfire_view view, bool shade, int titlebar_height)
+    {
+        if (!bool(enable_shade))
+        {
+            return;
+        }
+
+        if (shade)
+        {
+            if (view && view->is_mapped())
+            {
+                auto tr = ensure_transformer(view, titlebar_height);
+                tr->set_titlebar_height(titlebar_height);
+                tr->init_animation(shade);
+            }
+        } else
+        {
+            if (auto tr =
+                    view->get_transformed_node()->get_transformer<pixdecor_shade>(
+                        shade_transformer_name))
+            {
+                tr->set_titlebar_height(titlebar_height);
+                tr->init_animation(shade);
+            }
+        }
+    }
+
+    void handle_action(pixdecor_layout_t::action_response_t action)
     {
         if (auto view = _view.lock())
         {
             switch (action.action)
             {
-              case wf::pixdecor::DECORATION_ACTION_MOVE:
+              case DECORATION_ACTION_MOVE:
                 return wf::get_core().default_wm->move_request(view);
 
-              case wf::pixdecor::DECORATION_ACTION_RESIZE:
+              case DECORATION_ACTION_RESIZE:
                 return wf::get_core().default_wm->resize_request(view, action.edges);
 
-              case wf::pixdecor::DECORATION_ACTION_CLOSE:
+              case DECORATION_ACTION_CLOSE:
                 return view->close();
 
-              case wf::pixdecor::DECORATION_ACTION_TOGGLE_MAXIMIZE:
+              case DECORATION_ACTION_TOGGLE_MAXIMIZE:
                 if (view->pending_tiled_edges())
                 {
                     return wf::get_core().default_wm->tile_request(view, 0);
@@ -342,7 +449,15 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
 
                 break;
 
-              case wf::pixdecor::DECORATION_ACTION_MINIMIZE:
+              case DECORATION_ACTION_SHADE:
+                init_shade(view, true, current_titlebar);
+                break;
+
+              case DECORATION_ACTION_UNSHADE:
+                init_shade(view, false, current_titlebar);
+                break;
+
+              case DECORATION_ACTION_MINIMIZE:
                 return wf::get_core().default_wm->minimize_request(view, true);
                 break;
 
@@ -371,6 +486,17 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
         current_cursor_position = position;
     }
 
+    void recreate_frame()
+    {
+        update_decoration_size();
+        if (auto view = _view.lock())
+        {
+            auto size = wf::dimensions(view->get_pending_geometry());
+            layout.resize(size.width, size.height);
+            wf::get_core().tx_manager->schedule_object(view->toplevel());
+        }
+    }
+
     void resize(wf::dimensions_t dims)
     {
         if (auto view = _view.lock())
@@ -393,7 +519,9 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
     {
         if (auto view = _view.lock())
         {
+            view->damage();
             bool fullscreen = view->toplevel()->pending().fullscreen;
+            bool maximized  = view->toplevel()->pending().tiled_edges;
             if (fullscreen)
             {
                 current_thickness = 0;
@@ -401,19 +529,32 @@ class simple_decoration_node_t : public wf::scene::node_t, public wf::pointer_in
                 this->cached_region.clear();
             } else
             {
-                current_thickness = theme.get_border_size() +
-                    ((std::string(overlay_engine) == "rounded_corners") ? int(shadow_radius) * 2 : 0);
-                current_titlebar =
-                    theme.get_title_height() + current_thickness;
+                int shadow_thickness = std::string(overlay_engine) == "rounded_corners" &&
+                    (!maximized || (maximized && maximized_shadows)) ? int(shadow_radius) * 2 : 0;
+
+                current_thickness = theme.get_border_size() + shadow_thickness;
+                current_titlebar  = theme.get_title_height() +
+                    ((maximized && !titlebar_opt && !maximized_borders &&
+                        !maximized_shadows) ? 0 : current_thickness);
                 this->cached_region = layout.calculate_region();
             }
+
+            if (auto tr =
+                    view->get_transformed_node()->get_transformer<pixdecor_shade>(
+                        shade_transformer_name))
+            {
+                tr->set_titlebar_height(current_titlebar);
+            }
+
+            view->damage();
         }
     }
 };
 
-wf::simple_decorator_t::simple_decorator_t(wayfire_toplevel_view view)
+simple_decorator_t::simple_decorator_t(wayfire_toplevel_view view)
 {
     this->view = view;
+    this->shadow_thickness = 0;
     deco = std::make_shared<simple_decoration_node_t>(view);
     deco->resize(wf::dimensions(view->get_pending_geometry()));
     wf::scene::add_back(view->get_surface_root_node(), deco);
@@ -421,6 +562,7 @@ wf::simple_decorator_t::simple_decorator_t(wayfire_toplevel_view view)
     view->connect(&on_view_activated);
     view->connect(&on_view_geometry_changed);
     view->connect(&on_view_fullscreen);
+    view->connect(&on_view_tiled);
 
     on_view_activated = [this] (auto)
     {
@@ -430,15 +572,17 @@ wf::simple_decorator_t::simple_decorator_t(wayfire_toplevel_view view)
     on_view_geometry_changed = [this] (auto)
     {
         deco->resize(wf::dimensions(this->view->get_geometry()));
-        if (!this->view->toplevel()->pending().tiled_edges)
-        {
-            wf::get_core().tx_manager->schedule_object(this->view->toplevel());
-        }
+        wf::get_core().tx_manager->schedule_object(this->view->toplevel());
+    };
+
+    on_view_tiled = [this] (auto)
+    {
+        deco->resize(wf::dimensions(this->view->get_geometry()));
+        wf::get_core().tx_manager->schedule_object(this->view->toplevel());
     };
 
     on_view_fullscreen = [this] (auto)
     {
-        deco->update_decoration_size();
         if (!this->view->toplevel()->pending().fullscreen)
         {
             deco->resize(wf::dimensions(this->view->get_geometry()));
@@ -447,24 +591,38 @@ wf::simple_decorator_t::simple_decorator_t(wayfire_toplevel_view view)
     };
 }
 
-wf::simple_decorator_t::~simple_decorator_t()
+simple_decorator_t::~simple_decorator_t()
 {
     wf::scene::remove_child(deco);
     deco.reset();
 }
 
-void wf::simple_decorator_t::update_colors()
+int simple_decorator_t::get_titlebar_height()
+{
+    return deco->current_titlebar;
+}
+
+void simple_decorator_t::recreate_frame()
+{
+    deco->recreate_frame();
+}
+
+void simple_decorator_t::update_decoration_size()
+{
+    deco->update_decoration_size();
+}
+
+void simple_decorator_t::update_colors()
 {
     deco->theme.update_colors();
 }
 
-void wf::simple_decorator_t::effect_updated()
+void simple_decorator_t::effect_updated()
 {
-    deco->update_decoration_size();
     deco->theme.smoke.effect_updated();
 }
 
-wf::decoration_margins_t wf::simple_decorator_t::get_margins(const wf::toplevel_state_t& state)
+wf::decoration_margins_t simple_decorator_t::get_margins(const wf::toplevel_state_t& state)
 {
     if (state.fullscreen)
     {
@@ -473,16 +631,16 @@ wf::decoration_margins_t wf::simple_decorator_t::get_margins(const wf::toplevel_
 
     deco->theme.set_maximize(state.tiled_edges);
 
-    int thickness = deco->theme.get_border_size() +
-        ((std::string(overlay_engine) == "rounded_corners" &&
-            (!state.tiled_edges || maximized_shadows)) ? int(shadow_radius) * 2 : 0);
-    int titlebar = deco->theme.get_title_height() +
+    this->shadow_thickness = std::string(overlay_engine) == "rounded_corners" &&
+        (!state.tiled_edges || (state.tiled_edges && maximized_shadows)) ? int(shadow_radius) * 2 : 0;
+
+    int thickness = deco->theme.get_border_size() + this->shadow_thickness;
+    int titlebar  = deco->theme.get_title_height() +
         ((state.tiled_edges && !titlebar_opt && !maximized_borders && !maximized_shadows) ? 0 : thickness);
     if (state.tiled_edges && !maximized_borders)
     {
         if (maximized_shadows)
         {
-            thickness -= deco->theme.get_border_size();
             if (!titlebar_opt)
             {
                 titlebar = thickness;
@@ -493,8 +651,29 @@ wf::decoration_margins_t wf::simple_decorator_t::get_margins(const wf::toplevel_
         }
     }
 
-    deco->current_thickness = thickness;
-    deco->current_titlebar  = titlebar;
+    double shade_progress = 0.0;
+    if (auto tr =
+            view->get_transformed_node()->get_transformer<pixdecor_shade>(
+                shade_transformer_name))
+    {
+        tr->set_titlebar_height(titlebar);
+        shade_progress = tr->progression.shade;
+    }
+
+    if (view->has_data(custom_data_name))
+    {
+        view->get_data<wf_shadow_margin_t>(custom_data_name)->set_margins(
+            {shadow_thickness, shadow_thickness, shadow_thickness,
+                shadow_thickness +
+                int((view->get_geometry().height - shadow_thickness - titlebar) * shade_progress)});
+    } else
+    {
+        view->store_data(std::make_unique<wf_shadow_margin_t>(), custom_data_name);
+        view->get_data<wf_shadow_margin_t>(custom_data_name)->set_margins(
+            {shadow_thickness, shadow_thickness, shadow_thickness,
+                shadow_thickness +
+                int((view->get_geometry().height - shadow_thickness - titlebar) * shade_progress)});
+    }
 
     return wf::decoration_margins_t{
         .left   = thickness,
@@ -504,7 +683,7 @@ wf::decoration_margins_t wf::simple_decorator_t::get_margins(const wf::toplevel_
     };
 }
 
-void wf::simple_decorator_t::update_animation()
+void simple_decorator_t::update_animation()
 {
     auto margins = get_margins(view->toplevel()->current());
     auto bbox    = deco->get_bounding_box();
@@ -516,3 +695,5 @@ void wf::simple_decorator_t::update_animation()
     region |= wlr_box{bbox.x + bbox.width - margins.right, bbox.y, margins.right, bbox.height};
     wf::scene::damage_node(deco, region);
 }
+}
+}
