@@ -18,7 +18,14 @@
 #include <unistd.h>
 #include <gio/gio.h>
 #include <dlfcn.h>
+#include <wayfire/bindings-repository.hpp>
+#include <wayfire/plugins/ipc/ipc-activator.hpp>
+#include "shade.hpp"
 
+namespace wf
+{
+namespace pixdecor
+{
 int handle_theme_updated(int fd, uint32_t mask, void *data)
 {
     int bufsz = sizeof(inotify_event) + NAME_MAX + 1;
@@ -42,6 +49,8 @@ int handle_theme_updated(int fd, uint32_t mask, void *data)
 class wayfire_pixdecor : public wf::plugin_interface_t
 {
     wf::option_wrapper_t<int> border_size{"pixdecor/border_size"};
+    wf::option_wrapper_t<std::string> title_font{"pixdecor/title_font"};
+    wf::option_wrapper_t<int> title_text_align{"pixdecor/title_text_align"};
     wf::option_wrapper_t<bool> titlebar{"pixdecor/titlebar"};
     wf::option_wrapper_t<bool> maximized_borders{"pixdecor/maximized_borders"};
     wf::option_wrapper_t<bool> maximized_shadows{"pixdecor/maximized_shadows"};
@@ -49,6 +58,14 @@ class wayfire_pixdecor : public wf::plugin_interface_t
     wf::option_wrapper_t<wf::color_t> bg_color{"pixdecor/bg_color"};
     wf::option_wrapper_t<wf::color_t> fg_text_color{"pixdecor/fg_text_color"};
     wf::option_wrapper_t<wf::color_t> bg_text_color{"pixdecor/bg_text_color"};
+    wf::option_wrapper_t<wf::color_t> button_color{"pixdecor/button_color"};
+    wf::option_wrapper_t<double> button_line_thickness{"pixdecor/button_line_thickness"};
+    wf::option_wrapper_t<int> button_spacing{"pixdecor/button_spacing"};
+    wf::option_wrapper_t<int> button_x_offset{"pixdecor/button_x_offset"};
+    wf::option_wrapper_t<int> button_y_offset{"pixdecor/button_y_offset"};
+    wf::option_wrapper_t<std::string> button_minimize_image{"pixdecor/button_minimize_image"};
+    wf::option_wrapper_t<std::string> button_maximize_image{"pixdecor/button_maximize_image"};
+    wf::option_wrapper_t<std::string> button_close_image{"pixdecor/button_close_image"};
     wf::option_wrapper_t<std::string> ignore_views_string{"pixdecor/ignore_views"};
     wf::option_wrapper_t<std::string> always_decorate_string{"pixdecor/always_decorate"};
     wf::option_wrapper_t<std::string> effect_type{"pixdecor/effect_type"};
@@ -59,6 +76,10 @@ class wayfire_pixdecor : public wf::plugin_interface_t
     wf::option_wrapper_t<wf::color_t> shadow_color{"pixdecor/shadow_color"};
     wf::view_matcher_t ignore_views{"pixdecor/ignore_views"};
     wf::view_matcher_t always_decorate{"pixdecor/always_decorate"};
+    wf::option_wrapper_t<wf::keybinding_t> shade_modifier{"pixdecor/shade_modifier"};
+    wf::option_wrapper_t<int> csd_titlebar_height{"pixdecor/csd_titlebar_height"};
+    wf::option_wrapper_t<bool> enable_shade{"pixdecor/enable_shade"};
+    wf::ipc_activator_t pixdecor_toggle_shade{"pixdecor/shade_toggle"};
     wf::wl_idle_call idle_update_views;
     int inotify_fd;
     int wd_cfg_file;
@@ -66,9 +87,10 @@ class wayfire_pixdecor : public wf::plugin_interface_t
     wl_event_source *evsrc;
     std::function<void(void)> update_event;
     wf::effect_hook_t pre_hook;
-    wf::output_t *output;
     bool hook_set = false;
 
+    wf::axis_callback shade_axis_cb;
+
     wf::signal::connection_t<wf::txn::new_transaction_signal> on_new_tx =
         [=] (wf::txn::new_transaction_signal *ev)
     {
@@ -79,8 +101,9 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             {
                 // First check whether the toplevel already has decoration
                 // In that case, we should just set the correct margins
-                if (auto deco = toplevel->get_data<wf::simple_decorator_t>())
+                if (auto deco = toplevel->get_data<simple_decorator_t>())
                 {
+                    deco->update_decoration_size();
                     toplevel->pending().margins = deco->get_margins(toplevel->pending());
                     continue;
                 }
@@ -97,6 +120,11 @@ class wayfire_pixdecor : public wf::plugin_interface_t
                 wf::dassert(view != nullptr, "Mapping a toplevel means there must be a corresponding view!");
                 if (should_decorate_view(view))
                 {
+                    if (auto deco = toplevel->get_data<simple_decorator_t>())
+                    {
+                        deco->update_decoration_size();
+                    }
+
                     adjust_new_decorations(view);
                 }
             }
@@ -106,15 +134,202 @@ class wayfire_pixdecor : public wf::plugin_interface_t
     wf::signal::connection_t<wf::view_decoration_state_updated_signal> on_decoration_state_changed =
         [=] (wf::view_decoration_state_updated_signal *ev)
     {
+        auto toplevel = wf::toplevel_cast(ev->view);
+        if (toplevel)
+        {
+            remove_decoration(toplevel);
+        }
+
         update_view_decoration(ev->view);
     };
 
+    // allows criteria containing maximized or floating check
+    wf::signal::connection_t<wf::view_tiled_signal> on_view_tiled =
+        [=] (wf::view_tiled_signal *ev)
+    {
+        update_view_decoration(ev->view);
+    };
+
+    wf::signal::connection_t<wf::view_app_id_changed_signal> on_app_id_changed =
+        [=] (wf::view_app_id_changed_signal *ev)
+    {
+        update_view_decoration(ev->view);
+    };
+
+    wf::signal::connection_t<wf::view_title_changed_signal> on_title_changed =
+        [=] (wf::view_title_changed_signal *ev)
+    {
+        update_view_decoration(ev->view);
+    };
+
+    wf::signal::connection_t<wf::output_added_signal> on_output_added =
+        [=] (wf::output_added_signal *ev)
+    {
+        if (effect_animate || (std::string(effect_type) == "smoke") || (std::string(effect_type) == "ink"))
+        {
+            ev->output->render->add_effect(&pre_hook, wf::OUTPUT_EFFECT_PRE);
+        }
+    };
+
+    wf::signal::connection_t<wf::output_removed_signal> on_output_removed =
+        [=] (wf::output_removed_signal *ev)
+    {
+        ev->output->render->rem_effect(&pre_hook);
+    };
+
+    void pop_transformer(wayfire_view view)
+    {
+        if (view->get_transformed_node()->get_transformer(shade_transformer_name))
+        {
+            view->get_transformed_node()->rem_transformer(shade_transformer_name);
+        }
+    }
+
+    void remove_shade_transformers()
+    {
+        for (auto& view : wf::get_core().get_all_views())
+        {
+            pop_transformer(view);
+        }
+    }
+
+    std::shared_ptr<pixdecor_shade> ensure_transformer(wayfire_view view, int titlebar_height)
+    {
+        auto tmgr = view->get_transformed_node();
+        if (auto tr = tmgr->get_transformer<pixdecor_shade>(shade_transformer_name))
+        {
+            return tr;
+        }
+
+        auto node = std::make_shared<pixdecor_shade>(view, titlebar_height);
+        tmgr->add_transformer(node, wf::TRANSFORMER_2D, shade_transformer_name);
+        auto tr = tmgr->get_transformer<pixdecor_shade>(shade_transformer_name);
+
+        return tr;
+    }
+
+    void init_shade(wayfire_view view, bool shade, int titlebar_height)
+    {
+        if (!bool(enable_shade))
+        {
+            return;
+        }
+
+        if (shade)
+        {
+            if (view && view->is_mapped())
+            {
+                auto tr = ensure_transformer(view, titlebar_height);
+                tr->set_titlebar_height(titlebar_height);
+                tr->init_animation(shade);
+            }
+        } else
+        {
+            if (auto tr =
+                    view->get_transformed_node()->get_transformer<pixdecor_shade>(
+                        shade_transformer_name))
+            {
+                tr->set_titlebar_height(titlebar_height);
+                tr->init_animation(shade);
+            }
+        }
+    }
+
   public:
 
     void init() override
     {
+        auto& ol = wf::get_core().output_layout;
+        ol->connect(&on_output_added);
+        ol->connect(&on_output_removed);
         wf::get_core().connect(&on_decoration_state_changed);
         wf::get_core().tx_manager->connect(&on_new_tx);
+        wf::get_core().connect(&on_app_id_changed);
+        wf::get_core().connect(&on_title_changed);
+        wf::get_core().connect(&on_view_tiled);
+
+        if (bool(enable_shade))
+        {
+            wf::get_core().bindings->add_axis(shade_modifier, &shade_axis_cb);
+        }
+
+        pixdecor_toggle_shade.set_handler([=] (wf::output_t *output, wayfire_view view)
+        {
+            if (!bool(enable_shade))
+            {
+                return false;
+            }
+
+            if (auto toplevel = wf::toplevel_cast(view))
+            {
+                bool direction = true;
+                auto deco = toplevel->toplevel()->get_data<simple_decorator_t>();
+                if (auto tr =
+                        view->get_transformed_node()->get_transformer<pixdecor_shade>(
+                            shade_transformer_name))
+                {
+                    direction = !tr->get_direction();
+                }
+
+                init_shade(view, direction,
+                    deco ? deco->get_titlebar_height() : csd_titlebar_height);
+                return true;
+            }
+
+            return false;
+        });
+
+        shade_axis_cb = [=] (wlr_pointer_axis_event *ev)
+        {
+            auto v = wf::get_core().get_cursor_focus_view();
+            if (ev->orientation == WL_POINTER_AXIS_VERTICAL_SCROLL)
+            {
+                if (auto toplevel = wf::toplevel_cast(v))
+                {
+                    auto deco = toplevel->toplevel()->get_data<simple_decorator_t>();
+                    init_shade(v, ev->delta < 0 ? true : false,
+                        deco ? deco->get_titlebar_height() : csd_titlebar_height);
+                    return true;
+                }
+
+                return true;
+            }
+
+            return false;
+        };
+
+        enable_shade.set_callback([=]
+        {
+            if (bool(enable_shade))
+            {
+                wf::get_core().bindings->add_axis(shade_modifier, &shade_axis_cb);
+            }
+
+            {
+                wf::get_core().bindings->rem_binding(&shade_axis_cb);
+                remove_shade_transformers();
+            }
+        });
+
+        csd_titlebar_height.set_callback([=] ()
+        {
+            for (auto& view : wf::get_core().get_all_views())
+            {
+                if (auto tr =
+                        view->get_transformed_node()->get_transformer<pixdecor_shade>(
+                            shade_transformer_name))
+                {
+                    auto toplevel = toplevel_cast(view);
+                    if (toplevel)
+                    {
+                        if (!toplevel->toplevel()->get_data<simple_decorator_t>())
+                        {
+                            tr->set_titlebar_height(csd_titlebar_height);
+                        }
+                    }
+                }
+            }
+        });
 
         for (auto& view : wf::get_core().get_all_views())
         {
@@ -126,7 +341,7 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             for (auto& view : wf::get_core().get_all_views())
             {
                 auto toplevel = wf::toplevel_cast(view);
-                if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>())
+                if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>())
                 {
                     continue;
                 }
@@ -179,12 +394,12 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             for (auto& view : wf::get_core().get_all_views())
             {
                 auto toplevel = wf::toplevel_cast(view);
-                if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>())
+                if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>())
                 {
                     continue;
                 }
 
-                auto deco = toplevel->toplevel()->get_data<wf::simple_decorator_t>();
+                auto deco = toplevel->toplevel()->get_data<simple_decorator_t>();
                 deco->update_animation();
             }
         };
@@ -201,8 +416,24 @@ class wayfire_pixdecor : public wf::plugin_interface_t
 
         titlebar.set_callback([=] {option_changed_cb(false, true);});
         effect_type.set_callback([=] {option_changed_cb(false, false);});
-        overlay_engine.set_callback([=] {option_changed_cb(true, false);});
+        overlay_engine.set_callback([=] {option_changed_cb(true, false);recreate_frames();});
         effect_animate.set_callback([=] {option_changed_cb(false, false);});
+        button_color.set_callback([=] {recreate_frames();});
+        button_line_thickness.set_callback([=] {recreate_frames();});
+        button_spacing.set_callback([=] {recreate_frames();});
+        button_x_offset.set_callback([=] {recreate_frames();});
+        button_y_offset.set_callback([=] {recreate_frames();});
+        button_minimize_image.set_callback([=] {recreate_frames();});
+        button_maximize_image.set_callback([=] {recreate_frames();});
+        button_close_image.set_callback([=] {recreate_frames();});
+        title_text_align.set_callback([=]
+        {
+            for (auto& view : wf::get_core().get_all_views())
+            {
+                view->damage();
+            }
+        });
+        title_font.set_callback([=] {recreate_frames();});
         shadow_radius.set_callback([=]
         {
             option_changed_cb(false, (std::string(overlay_engine) == "rounded_corners"));
@@ -212,7 +443,7 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             for (auto& view : wf::get_core().get_all_views())
             {
                 auto toplevel = wf::toplevel_cast(view);
-                if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>())
+                if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>())
                 {
                     continue;
                 }
@@ -225,7 +456,7 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             for (auto& view : wf::get_core().get_all_views())
             {
                 auto toplevel = wf::toplevel_cast(view);
-                if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>())
+                if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>())
                 {
                     continue;
                 }
@@ -239,7 +470,7 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             for (auto& view : wf::get_core().get_all_views())
             {
                 auto toplevel = wf::toplevel_cast(view);
-                if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>() ||
+                if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>() ||
                     !toplevel->pending_tiled_edges())
                 {
                     continue;
@@ -256,7 +487,7 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             for (auto& view : wf::get_core().get_all_views())
             {
                 auto toplevel = wf::toplevel_cast(view);
-                if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>() ||
+                if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>() ||
                     !toplevel->pending_tiled_edges())
                 {
                     continue;
@@ -309,12 +540,47 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             }
         }
 
+        on_output_added.disconnect();
+        on_output_removed.disconnect();
+        on_decoration_state_changed.disconnect();
+        on_new_tx.disconnect();
+        on_app_id_changed.disconnect();
+        on_title_changed.disconnect();
+
+        wf::get_core().bindings->rem_binding(&shade_axis_cb);
+        remove_shade_transformers();
+
         wl_event_source_remove(evsrc);
         inotify_rm_watch(inotify_fd, wd_cfg_file);
         inotify_rm_watch(inotify_fd, wd_cfg_dir);
         close(inotify_fd);
     }
 
+    void recreate_frame(wayfire_toplevel_view toplevel)
+    {
+        auto deco = toplevel->toplevel()->get_data<simple_decorator_t>();
+        if (!deco)
+        {
+            return;
+        }
+
+        deco->recreate_frame();
+    }
+
+    void recreate_frames()
+    {
+        for (auto& view : wf::get_core().get_all_views())
+        {
+            auto toplevel = wf::toplevel_cast(view);
+            if (!toplevel)
+            {
+                continue;
+            }
+
+            recreate_frame(toplevel);
+        }
+    }
+
     void option_changed_cb(bool resize_decorations, bool recreate_decorations)
     {
         if (effect_animate || (std::string(effect_type) == "smoke") || (std::string(effect_type) == "ink"))
@@ -346,7 +612,7 @@ class wayfire_pixdecor : public wf::plugin_interface_t
             for (auto& view : wf::get_core().get_all_views())
             {
                 auto toplevel = wf::toplevel_cast(view);
-                if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>())
+                if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>())
                 {
                     continue;
                 }
@@ -362,13 +628,13 @@ class wayfire_pixdecor : public wf::plugin_interface_t
         for (auto& view : wf::get_core().get_all_views())
         {
             auto toplevel = wf::toplevel_cast(view);
-            if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>())
+            if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>())
             {
                 continue;
             }
 
             view->damage();
-            toplevel->toplevel()->get_data<wf::simple_decorator_t>()->effect_updated();
+            toplevel->toplevel()->get_data<simple_decorator_t>()->effect_updated();
 
             auto& pending = toplevel->toplevel()->pending();
             if (!resize_decorations || (pending.tiled_edges != 0))
@@ -385,12 +651,10 @@ class wayfire_pixdecor : public wf::plugin_interface_t
                 pending.geometry = wf::expand_geometry_by_margins(pending.geometry, pending.margins);
             } else
             {
-                pending.margins =
-                {int(shadow_radius) * 2, int(shadow_radius) * 2,
-                    int(shadow_radius) * 2, int(shadow_radius) * 2};
                 pending.geometry = wf::shrink_geometry_by_margins(pending.geometry, pending.margins);
-                pending.margins  = toplevel->toplevel()->get_data<wf::simple_decorator_t>()->get_margins(
+                pending.margins  = toplevel->toplevel()->get_data<simple_decorator_t>()->get_margins(
                     toplevel->toplevel()->pending());
+                pending.geometry = wf::expand_geometry_by_margins(pending.geometry, pending.margins);
             }
 
             wf::get_core().tx_manager->schedule_object(toplevel->toplevel());
@@ -402,12 +666,12 @@ class wayfire_pixdecor : public wf::plugin_interface_t
         for (auto& view : wf::get_core().get_all_views())
         {
             auto toplevel = wf::toplevel_cast(view);
-            if (!toplevel || !toplevel->toplevel()->get_data<wf::simple_decorator_t>())
+            if (!toplevel || !toplevel->toplevel()->get_data<simple_decorator_t>())
             {
                 continue;
             }
 
-            auto deco = toplevel->toplevel()->get_data<wf::simple_decorator_t>();
+            auto deco = toplevel->toplevel()->get_data<simple_decorator_t>();
             deco->update_colors();
             view->damage();
         }
@@ -435,12 +699,12 @@ class wayfire_pixdecor : public wf::plugin_interface_t
     {
         auto toplevel = view->toplevel();
 
-        if (!toplevel->get_data<wf::simple_decorator_t>())
+        if (!toplevel->get_data<simple_decorator_t>())
         {
-            toplevel->store_data(std::make_unique<wf::simple_decorator_t>(view));
+            toplevel->store_data(std::make_unique<simple_decorator_t>(view));
         }
 
-        auto deco     = toplevel->get_data<wf::simple_decorator_t>();
+        auto deco     = toplevel->get_data<simple_decorator_t>();
         auto& pending = toplevel->pending();
         pending.margins = deco->get_margins(pending);
 
@@ -453,7 +717,7 @@ class wayfire_pixdecor : public wf::plugin_interface_t
 
     void remove_decoration(wayfire_toplevel_view view)
     {
-        view->toplevel()->erase_data<wf::simple_decorator_t>();
+        view->toplevel()->erase_data<simple_decorator_t>();
         auto& pending = view->toplevel()->pending();
         if (!pending.fullscreen && !pending.tiled_edges)
         {
@@ -461,6 +725,12 @@ class wayfire_pixdecor : public wf::plugin_interface_t
         }
 
         pending.margins = {0, 0, 0, 0};
+
+        std::string custom_data_name = "wf-decoration-shadow-margin";
+        if (view->has_data(custom_data_name))
+        {
+            view->erase_data(custom_data_name);
+        }
     }
 
     void update_view_decoration(wayfire_view view)
@@ -472,10 +742,10 @@ class wayfire_pixdecor : public wf::plugin_interface_t
                 return;
             }
 
-            if (!toplevel->toplevel()->get_data<wf::simple_decorator_t>() && (always_decorate.matches(view) ||
-                                                                              (should_decorate_view(toplevel)
-                                                                               &&
-                                                                               !ignore_views.matches(view))))
+            if (!toplevel->toplevel()->get_data<simple_decorator_t>() && (always_decorate.matches(view) ||
+                                                                          (should_decorate_view(toplevel)
+                                                                           &&
+                                                                           !ignore_views.matches(view))))
             {
                 adjust_new_decorations(toplevel);
             } else if ((!always_decorate.matches(view) &&
@@ -488,5 +758,7 @@ class wayfire_pixdecor : public wf::plugin_interface_t
         }
     }
 };
+}
+}
 
-DECLARE_WAYFIRE_PLUGIN(wayfire_pixdecor);
+DECLARE_WAYFIRE_PLUGIN(wf::pixdecor::wayfire_pixdecor);
